@@ -300,6 +300,13 @@ find_varname(Var, [_|T], Name) :-
 %   This predicate calls the multifile predicate unify_clause_hook/5
 %   with the same arguments to support user extensions.
 %
+%   @arg Module is the source module that   was active when loading this
+%   clause,  which  is  the  same  as  prolog_load_context/2  using  the
+%   `module` context. If this cannot be established  it is the module to
+%   which the clause itself is associated.   The argument may be used to
+%   determine whether or not a specific user transformation is in scope.
+%   See also term_expansion/2,4 and goal_expansion/2,4.
+%
 %   @tbd    This really must be  more   flexible,  dealing with much
 %           more complex source-translations,  falling   back  to  a
 %           heuristic method locating as much as possible.
@@ -308,6 +315,14 @@ unify_clause(Read, _, _, _, _) :-
     var(Read),
     !,
     fail.
+unify_clause((RHead :- RBody), (CHead :- CBody), Module, TermPos1, TermPos) :-
+    '$expand':f2_pos(TermPos1, HPos, BPos1,
+                     TermPos2, HPos, BPos2),
+    inlined_unification(RBody, CBody, RBody1, CBody1, RHead,
+                        BPos1, BPos2),
+    RBody1 \== RBody,
+    !,
+    unify_clause((RHead :- RBody1), (CHead :- CBody1), Module, TermPos2, TermPos).
 unify_clause(Read, Decompiled, _, TermPos, TermPos) :-
     Read =@= Decompiled,
     !,
@@ -358,28 +373,28 @@ unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
     !,
     TermPos2 = term_position(F,T,FF,FT,[ HP, BP ]),
     match_module(Compiled2, Compiled1, Module, TermPos2, TermPos).
-unify_clause((Head,Cond => Body), Compiled1, Module,
+unify_clause((Head,RCond => Body), (CHead :- CCondAndBody), Module,
              term_position(F,T,FF,FT,
                            [ term_position(_,_,_,_,[HP,CP]),
                              BP
                            ]),
              TermPos) :-
+    split_on_cut(CCondAndBody, CCond, CBody),
     !,
-    TermPos1 = term_position(F,T,FF,FT,
-                             [ HP,
-                               term_position(_,_,_,_,
-                                             [ CP,
-                                               term_position(_,_,_,_,
-                                                             [ FF-FT,
-                                                               BP
-                                                             ])
-                                             ])
-                             ]),
-    unify_clause((Head :- Cond, !, Body), Compiled1, Module, TermPos1, TermPos).
+    inlined_unification(RCond, CCond, RCond1, CCond1, Head, CP, CP1),
+    TermPos1 = term_position(F,T,FF,FT, [HP, BP1]),
+    BP2 = term_position(_,_,_,_, [FF-FT, BP]), % Represent (!, Body), placing
+    (   CCond1 == true                         % ! at =>
+    ->  BP1 = BP2,                             % Whole guard is inlined
+        unify_clause((Head :- !, Body), (CHead :- !, CBody),
+                     Module, TermPos1, TermPos)
+    ;   BP1 = term_position(_,_,_,_, [CP1, BP2]),
+        unify_clause((Head :- RCond1, !, Body), (CHead :- CCond1, !, CBody),
+                     Module, TermPos1, TermPos)
+    ).
 unify_clause((Head => Body), Compiled1, Module, TermPos0, TermPos) :-
     !,
     unify_clause(Head :- Body, Compiled1, Module, TermPos0, TermPos).
-                                        % general term-expansion
 unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
     ci_expand(Read, Compiled2, Module, TermPos0, TermPos1),
     match_module(Compiled2, Compiled1, Module, TermPos1, TermPos).
@@ -392,6 +407,58 @@ unify_clause_head(H1, H2) :-
     strip_module(H1, _, H),
     strip_module(H2, _, H).
 
+%!  inlined_unification(+BodyRead, +BodyCompiled,
+%!                      -BodyReadOut, -BodyCompiledOut,
+%!                      +HeadRead,
+%!                      +BodyPosIn, -BodyPosOut) is det.
+
+inlined_unification((V=T,RBody0), (CV=CT,CBody0),
+                    RBody, CBody, RHead, BPos1, BPos),
+    sub_term(V2, RHead),
+    V == V2,
+    (V=T) =@= (CV=CT) =>
+    argpos(2, BPos1, BPos2),
+    inlined_unification(RBody0, CBody0, RBody, CBody, RHead, BPos2, BPos).
+inlined_unification((V=T), (CV=CT),
+                    RBody, CBody, RHead, BPos1, BPos),
+    sub_term(V2, RHead),
+    V == V2,
+    (V=T) =@= (CV=CT) =>
+    RBody = true,
+    CBody = true,
+    argpos(2, BPos1, BPos).
+inlined_unification((V=T,RBody0), CBody0,
+                    RBody, CBody, RHead, BPos1, BPos),
+    sub_term(V2, RHead),
+    V == V2,
+    \+ (CBody0 = (G1,_), G1 \=@= (V=T)) =>
+    argpos(2, BPos1, BPos2),
+    inlined_unification(RBody0, CBody0, RBody, CBody, RHead, BPos2, BPos).
+inlined_unification((V=_), true,
+                    RBody, CBody, RHead, BPos1, BPos),
+    sub_term(V2, RHead),
+    V == V2 =>
+    RBody = true,
+    CBody = true,
+    argpos(2, BPos1, BPos).
+inlined_unification(RBody0, CBody0, RBody, CBody, _RHead,
+                    BPos0, BPos) =>
+    RBody = RBody0,
+    BPos  = BPos0,
+    CBody = CBody0.
+
+split_on_cut((Cond0,!,Body0), Cond, Body) =>
+    Cond = Cond0,
+    Body = Body0.
+split_on_cut((!,Body0), Cond, Body) =>
+    Cond = true,
+    Body = Body0.
+split_on_cut((A,B), Cond, Body) =>
+    Cond = (A,Cond1),
+    split_on_cut(B, Cond1, Body).
+split_on_cut(_, _, _) =>
+    fail.
+
 ci_expand(Read, Compiled, Module, TermPos0, TermPos) :-
     catch(setup_call_cleanup(
               ( set_xref_flag(OldXRef),
@@ -551,6 +618,8 @@ ubody(Read, Decompiled, Module, TermPosRead, TermPosDecompiled) :-
 ubody(X0, X, M,
       term_position(F,T,FF,TT,PA0),
       term_position(F,T,FF,TT,PA)) :-
+    callable(X0),
+    callable(X),
     meta(M, X0, S),
     !,
     X0 =.. [_|A0],
@@ -635,6 +704,16 @@ mkconj(Conj, M, term_position(0,0,0,0,[PA,PB]), GL, TG, PL, TP) :-
 mkconj(A0, M, P0, [A|TG], TG, [P|TP], TP) :-
     ubody(A, A0, M, P, P0).
 
+%!  argpos(+N, +PositionTerm, -ArgPositionTerm) is det.
+%
+%   Get the position for the nth argument of PositionTerm.
+
+argpos(N, parentheses_term_position(_,_,PosIn), Pos) =>
+    argpos(N, PosIn, Pos).
+argpos(N, term_position(_,_,_,_,ArgPos), Pos) =>
+    nth1(N, ArgPos, Pos).
+argpos(_, _, _) => true.
+
 
                  /*******************************
                  *    PCE STUFF (SHOULD MOVE)   *
