@@ -1101,6 +1101,7 @@ int
 nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr, term_t *varport)
 { int port;
 
+  memset(addr, 0, sizeof(*addr));
   addr->sin_family = AF_INET;
   addr->sin_addr.s_addr = INADDR_ANY;
 
@@ -1233,9 +1234,12 @@ nbio_connect(nbio_sock_t socket,
 #endif
 	continue;
       }
-#ifdef __WINDOWS__
+#if defined(WSAEISCONN)				 /* Windows */
       if ( GET_ERRNO == WSAEISCONN )
         break;
+#elif defined(EISCONN)
+      if ( GET_ERRNO == EISCONN )
+        break;
 #endif
       nbio_error(GET_ERRNO, TCP_ERRNO);
       return -1;
