@@ -446,7 +446,9 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	return;
       }
       if ( --me == bt->current-1 )
+      { Sdprintf("No backtrace named %s\n", why);
 	break;
+      }
     }
   }
 }
@@ -1014,7 +1016,7 @@ initBackTrace(void)
 {
 }
 
-#ifdef HAVE_SIGNAL
+#if O_SIGNALS && defined(HAVE_SIGNAL)
 void
 sigCrashHandler(int sig)
 { int tid;
@@ -1050,6 +1052,6 @@ sigCrashHandler(int sig)
 { fatalError("Something went wrong");
 }
 
-#endif /*HAVE_SIGNAL*/
+#endif /*O_SIGNALS && HAVE_SIGNAL*/
 
 #endif /*BTRACE_DONE*/
