@@ -51,6 +51,10 @@
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
 #endif
+#ifdef HAVE_TCMALLOC_EXTENSION_C_H
+/* Provides error checking for the weak declarations below */
+#include <gperftools/malloc_extension_c.h>
+#endif
 #ifdef HAVE_SYS_MMAN_H
 #define MMAP_STACK 1
 #include <sys/mman.h>
@@ -323,14 +327,15 @@ typedef struct cycle_entry
 { LocalFrame frame;
 } cycle_entry;
 
+#define is_variant_frame(fr1, fr2) LDFUNC(is_variant_frame, fr1, fr2)
 static int
-is_variant_frame(LocalFrame fr1, LocalFrame fr2 ARG_LD)
+is_variant_frame(DECL_LD LocalFrame fr1, LocalFrame fr2)
 { if ( fr1->predicate == fr2->predicate )
   { size_t arity = fr1->predicate->functor->arity;
     size_t i;
 
     for(i=0; i<arity; i++)
-    { if ( !is_variant_ptr(argFrameP(fr1, i), argFrameP(fr2, i) PASS_LD) )
+    { if ( !is_variant_ptr(argFrameP(fr1, i), argFrameP(fr2, i)) )
 	return FALSE;
     }
 
@@ -341,11 +346,11 @@ is_variant_frame(LocalFrame fr1, LocalFrame fr2 ARG_LD)
 }
 
 
+#define non_terminating_recursion(fr0, ce, is_cycle) LDFUNC(non_terminating_recursion, fr0, ce, is_cycle)
 static int
-non_terminating_recursion(LocalFrame fr0,
+non_terminating_recursion(DECL_LD LocalFrame fr0,
 			  cycle_entry ce[MAX_CYCLE],
-			  int *is_cycle
-			  ARG_LD)
+			  int *is_cycle)
 { int depth, mindepth = 1, repeat;
   LocalFrame fr, ctx;
 
@@ -363,7 +368,7 @@ again:
   if ( !fr || depth >= MAX_CYCLE )
     return 0;
 
-  *is_cycle = is_variant_frame(fr0, fr PASS_LD);
+  *is_cycle = is_variant_frame(fr0, fr);
   ctx = fr;
 
   for(repeat=MIN_REPEAT; fr && --repeat > 0; )
@@ -392,15 +397,16 @@ again:
   return 0;
 }
 
+#define find_non_terminating_recursion(fr, ce, is_cycle) LDFUNC(find_non_terminating_recursion, fr, ce, is_cycle)
 static int
-find_non_terminating_recursion(LocalFrame fr, cycle_entry ce[MAX_CYCLE],
-			       int *is_cycle ARG_LD)
+find_non_terminating_recursion(DECL_LD LocalFrame fr, cycle_entry ce[MAX_CYCLE],
+			       int *is_cycle)
 { int max_pre_loop = MAX_PRE_LOOP;
 
   for(; fr && max_pre_loop; fr = parentFrame(fr), max_pre_loop--)
   { int len;
 
-    if ( (len=non_terminating_recursion(fr, ce, is_cycle PASS_LD)) )
+    if ( (len=non_terminating_recursion(fr, ce, is_cycle)) )
       return len;
   }
 
@@ -408,8 +414,9 @@ find_non_terminating_recursion(LocalFrame fr, cycle_entry ce[MAX_CYCLE],
 }
 
 
+#define top_of_stack(fr, ce, maxdepth) LDFUNC(top_of_stack, fr, ce, maxdepth)
 static int
-top_of_stack(LocalFrame fr, cycle_entry ce[MAX_CYCLE], int maxdepth ARG_LD)
+top_of_stack(DECL_LD LocalFrame fr, cycle_entry ce[MAX_CYCLE], int maxdepth)
 { int depth;
 
   for(depth = 0; fr && depth < maxdepth; fr = parentFrame(fr), depth++)
@@ -483,7 +490,7 @@ push_goal(LocalFrame fr)
 
 	if ( isList(*a) )
 	{ Word tail;
-	  intptr_t len = skip_list(a, &tail PASS_LD);
+	  intptr_t len = skip_list(a, &tail);
 
 	  *ad++ = FUNCTOR_dot2;
 	  deRef(tail);
@@ -548,8 +555,9 @@ push_cycle(cycle_entry ce[MAX_CYCLE], int depth)
 }
 
 
+#define push_stack(ce, depth, name, pp) LDFUNC(push_stack, ce, depth, name, pp)
 static void
-push_stack(cycle_entry ce[MAX_CYCLE], int depth, atom_t name, Word *pp ARG_LD)
+push_stack(DECL_LD cycle_entry ce[MAX_CYCLE], int depth, atom_t name, Word *pp)
 { word w;
   Word p = *pp;
 
@@ -601,19 +609,19 @@ push_overflow_context(Stack stack, int extra)
     { int is_cycle;
 
       if ( (depth=find_non_terminating_recursion(environment_frame, ce,
-						 &is_cycle PASS_LD)) )
+						 &is_cycle)) )
       { push_stack(ce, depth, is_cycle ? ATOM_cycle : ATOM_non_terminating,
-		   &p PASS_LD);
-      } else if ( (depth=top_of_stack(environment_frame, ce, 5 PASS_LD)) )
-      { push_stack(ce, depth, ATOM_stack, &p PASS_LD);
+		   &p);
+      } else if ( (depth=top_of_stack(environment_frame, ce, 5)) )
+      { push_stack(ce, depth, ATOM_stack, &p);
       }
-    } else if ( (depth=top_of_stack(environment_frame, ce, 5 PASS_LD)) )
-    { push_stack(ce, depth, ATOM_stack, &p PASS_LD);
+    } else if ( (depth=top_of_stack(environment_frame, ce, 5)) )
+    { push_stack(ce, depth, ATOM_stack, &p);
     }
 
     *dict = dict_functor((p-dict-2)/2);		/* final functor */
 
-    dict_order(dict, NULL PASS_LD);
+    dict_order(dict, NULL);
 
     return consPtr(dict, STG_GLOBAL|TAG_COMPOUND);
   } else
@@ -695,7 +703,7 @@ outOfStack(void *stack, stack_overflow_action how)
 	gTop += 5;
 
 	*valTermRef(LD->exception.bin) = consPtr(p, TAG_COMPOUND|STG_GLOBAL);
-	freezeGlobal(PASS_LD1);
+	freezeGlobal();
       } else
       { Sdprintf("ERROR: Out of global-stack.\n"
 		 "ERROR: No room for exception term.  Aborting.\n");
@@ -743,7 +751,7 @@ raiseStackOverflow(int overflow)
 
 
 void
-pushArgumentStack__LD(Word p ARG_LD)
+f_pushArgumentStack(DECL_LD Word p)
 { Word *newbase;
   size_t newsize = nextStackSize((Stack)&LD->stacks.argument, 1);
 
@@ -784,7 +792,7 @@ inline  as  it is simple and usualy very time critical.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-allocGlobal__LD(size_t n ARG_LD)
+allocGlobal(DECL_LD size_t n)
 { Word result;
 
   if ( !hasGlobalSpace(n) )
@@ -803,7 +811,7 @@ allocGlobal__LD(size_t n ARG_LD)
 }
 
 Word
-allocGlobalNoShift__LD(size_t n ARG_LD)
+allocGlobalNoShift(DECL_LD size_t n)
 { Word result;
 
   if ( gTop+n > gMax )
@@ -842,7 +850,7 @@ Return is one of:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-put_int64(Word at, int64_t l, int flags ARG_LD)
+put_int64(DECL_LD Word at, int64_t l, int flags)
 { Word p;
   word r, m;
   int req;
@@ -914,7 +922,7 @@ Note that these functions can trigger GC
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-allocString(size_t len ARG_LD)
+allocString(DECL_LD size_t len)
 { size_t lw = (len+sizeof(word))/sizeof(word);
   int pad = (int)(lw*sizeof(word) - len);
   Word p = allocGlobal(2 + lw);
@@ -934,7 +942,7 @@ allocString(size_t len ARG_LD)
 word
 globalString(size_t len, const char *s)
 { GET_LD
-  Word p = allocString(len+1 PASS_LD);
+  Word p = allocString(len+1);
 
   if ( p )
   { char *q = (char *)&p[1];
@@ -964,7 +972,7 @@ globalWString(size_t len, const pl_wchar_t *s)
   if ( p == e )				/* 8-bit string */
   { unsigned char *t;
 
-    if ( !(g = allocString(len+1 PASS_LD)) )
+    if ( !(g = allocString(len+1)) )
       return 0;
     t = (unsigned char *)&g[1];
     *t++ = 'B';
@@ -974,7 +982,7 @@ globalWString(size_t len, const pl_wchar_t *s)
   { char *t;
     pl_wchar_t *w;
 
-    if ( !(g = allocString((len+1)*sizeof(pl_wchar_t) PASS_LD)) )
+    if ( !(g = allocString((len+1)*sizeof(pl_wchar_t))) )
       return 0;
     t = (char *)&g[1];
     w = (pl_wchar_t*)t;
@@ -988,7 +996,7 @@ globalWString(size_t len, const pl_wchar_t *s)
 
 
 char *
-getCharsString__LD(word w, size_t *len ARG_LD)
+getCharsString(DECL_LD word w, size_t *len)
 { Word p = valPtr(w);
   word m = *p;
   size_t wn  = wsizeofInd(m);
@@ -1009,7 +1017,7 @@ getCharsString__LD(word w, size_t *len ARG_LD)
 
 
 pl_wchar_t *
-getCharsWString__LD(word w, size_t *len ARG_LD)
+getCharsWString(DECL_LD word w, size_t *len)
 { Word p = valPtr(w);
   word m = *p;
   size_t wn  = wsizeofInd(m);
@@ -1041,7 +1049,7 @@ might not be properly aligned.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-put_double(Word at, double d, int flags ARG_LD)
+put_double(DECL_LD Word at, double d, int flags)
 { Word p;
   word m = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
 
@@ -1068,31 +1076,7 @@ put_double(Word at, double d, int flags ARG_LD)
 }
 
 
-		 /*******************************
-		 *	  64-BIT INTEGERS	*
-		 *******************************/
-
-#if ALIGNOF_INT64_T != ALIGNOF_VOIDP
-
-int64_t					/* take care of alignment! */
-valBignum__LD(word w ARG_LD)
-{ Word p = valIndirectP(w);
-  union
-  { int64_t i;
-    word w[WORDS_PER_INT64];
-  } val;
-
-#if ( SIZEOF_VOIDP == 4 )
-  val.w[0] = p[0];
-  val.w[1] = p[1];
-#else
-#error "Unsupported int64_t alignment conversion"
-#endif
-
-  return val.i;
-}
-
-#endif
+/* valBignum(DECL_LD word w) moved to pl-inline.h */
 
 		 /*******************************
 		 *  GENERIC INDIRECT OPERATIONS	*
@@ -1121,7 +1105,7 @@ equalIndirect(word w1, word w2)
 word
 globalIndirectFromCode(Code *PC)
 { GET_LD
-  struct word_and_Code retval = VM_globalIndirectFromCode(*PC PASS_LD);
+  struct word_and_Code retval = VM_globalIndirectFromCode(*PC);
   *PC = retval.code;
   return retval.word;
 }
@@ -1658,12 +1642,12 @@ stack_nrealloc(void *mem, size_t req)
 		 *	       TCMALLOC		*
 		 *******************************/
 
-static int (*fMallocExtension_GetNumericProperty)(const char *, size_t *);
-static int (*fMallocExtension_SetNumericProperty)(const char *, size_t);
-static void (*fMallocExtension_MarkThreadIdle)(void) = NULL;
-static void (*fMallocExtension_MarkThreadTemporarilyIdle)(void) = NULL;
-static void (*fMallocExtension_MarkThreadBusy)(void) = NULL;
-static void (*fMallocExtension_ReleaseFreeMemory)(void) = NULL;
+WEAK_DECLARE(int, MallocExtension_GetNumericProperty, (const char *, size_t *));
+WEAK_DECLARE(int, MallocExtension_SetNumericProperty, (const char *, size_t));
+WEAK_DECLARE(void, MallocExtension_MarkThreadIdle, (void));
+WEAK_DECLARE(void, MallocExtension_MarkThreadTemporarilyIdle, (void));
+WEAK_DECLARE(void, MallocExtension_MarkThreadBusy, (void));
+WEAK_DECLARE(void, MallocExtension_ReleaseFreeMemory, (void));
 
 static const char* tcmalloc_properties[] =
 { "generic.current_allocated_bytes",
@@ -1698,7 +1682,7 @@ malloc_property(term_t prop, control_t handle)
 	  { if ( streq(s, *pname) )
 	    { size_t val;
 
-	      if ( fMallocExtension_GetNumericProperty(*pname, &val) )
+	      if ( WEAK_FUNC(MallocExtension_GetNumericProperty)(*pname, &val) )
 	      { term_t a = PL_new_term_ref();
 		_PL_get_arg(1, prop, a);
 		return PL_unify_uint64(a, val);
@@ -1723,7 +1707,7 @@ malloc_property(term_t prop, control_t handle)
       for(; *pname; pname++)
       { size_t val;
 
-	if ( fMallocExtension_GetNumericProperty(*pname, &val) )
+	if ( WEAK_FUNC(MallocExtension_GetNumericProperty)(*pname, &val) )
 	{ if ( PL_unify_term(prop, PL_FUNCTOR_CHARS, *pname, 1,
 			             PL_INT64, val) )
 	  { PL_close_foreign_frame(fid);
@@ -1773,7 +1757,7 @@ set_malloc(term_t prop)
 
       for(; *pname; pname++)
       { if ( streq(s, *pname) )
-	{ if ( fMallocExtension_SetNumericProperty(*pname, val) )
+	{ if ( WEAK_FUNC(MallocExtension_SetNumericProperty)(*pname, val) )
 	    return TRUE;
 	  else
 	    return PL_permission_error("set", "malloc_property", prop);
@@ -1792,8 +1776,7 @@ size_t
 heapUsed(void)
 { size_t val;
 
-  if (fMallocExtension_GetNumericProperty &&
-      fMallocExtension_GetNumericProperty("generic.current_allocated_bytes", &val))
+  if (WEAK_TRY_CALL(MallocExtension_GetNumericProperty, "generic.current_allocated_bytes", &val))
   {
 #ifdef MMAP_STACK
     val += GD->statistics.stack_space;
@@ -1826,11 +1809,10 @@ initTCMalloc(void)
     return is_tcmalloc;
   done = TRUE;
 
-  if ( (fMallocExtension_GetNumericProperty =
-		PL_dlsym(NULL, "MallocExtension_GetNumericProperty")) )
+  if ( WEAK_IMPORT(MallocExtension_GetNumericProperty) )
   { size_t in_use;
 
-    if ( fMallocExtension_GetNumericProperty("generic.current_allocated_bytes", &in_use) &&
+    if ( WEAK_FUNC(MallocExtension_GetNumericProperty)("generic.current_allocated_bytes", &in_use) &&
 	 in_use > 100000 )
     { is_tcmalloc = TRUE;
       PL_set_prolog_flag("malloc", PL_ATOM, "tcmalloc");
@@ -1842,28 +1824,26 @@ initTCMalloc(void)
 			PL_FA_NONDETERMINISTIC);
     set++;
   }
-  if ( (fMallocExtension_SetNumericProperty =
-		PL_dlsym(NULL, "MallocExtension_SetNumericProperty")) )
+  if ( WEAK_IMPORT(MallocExtension_SetNumericProperty) )
   { PL_register_foreign_in_module("system", "set_malloc", 1, set_malloc, 0);
     set++;
   }
 
-  fMallocExtension_MarkThreadIdle =
-    PL_dlsym(NULL, "MallocExtension_MarkThreadIdle");
-  fMallocExtension_MarkThreadTemporarilyIdle =
-    PL_dlsym(NULL, "MallocExtension_MarkThreadTemporarilyIdle");
-  fMallocExtension_MarkThreadBusy =
-    PL_dlsym(NULL, "MallocExtension_MarkThreadBusy");
-  fMallocExtension_ReleaseFreeMemory =
-    PL_dlsym(NULL, "MallocExtension_ReleaseFreeMemory");
+  WEAK_IMPORT(MallocExtension_MarkThreadIdle);
+  WEAK_IMPORT(MallocExtension_MarkThreadTemporarilyIdle);
+  WEAK_IMPORT(MallocExtension_MarkThreadBusy);
+  WEAK_IMPORT(MallocExtension_ReleaseFreeMemory);
 
   return set;
 }
 
-#ifdef HAVE_MALLINFO
 static int is_ptmalloc = FALSE;
-static struct mallinfo (*fmallinfo)(void) = NULL;
-static int             (*fmalloc_trim)(int pad) = NULL;
+#ifdef HAVE_MALLINFO2
+WEAK_DECLARE(struct mallinfo2, mallinfo2, (void));
+#elif defined(HAVE_MALLINFO)
+WEAK_DECLARE(struct mallinfo, mallinfo, (void));
+#endif
+WEAK_DECLARE(int, malloc_trim, (size_t pad));
 
 static int
 initPTMalloc(void)
@@ -1873,23 +1853,26 @@ initPTMalloc(void)
     return is_ptmalloc;
   done = TRUE;
 
-  if ( (fmallinfo    = PL_dlsym(NULL, "mallinfo")) &&
-       (fmalloc_trim = PL_dlsym(NULL, "malloc_trim")) )
-  { struct mallinfo info = fmallinfo();
-
-    if ( info.uordblks > 100000 )
-    { PL_set_prolog_flag("malloc", PL_ATOM, "ptmalloc");
-      is_ptmalloc = TRUE;
-    }
+  size_t uordblks = 0;
+#ifdef HAVE_MALLINFO2
+  if ( WEAK_IMPORT(mallinfo2) )
+  { uordblks = WEAK_FUNC(mallinfo2)().uordblks;
   }
+#elif defined(HAVE_MALLINFO)
+  if ( WEAK_IMPORT(mallinfo) && uordblks < 100000 )
+  { uordblks = WEAK_FUNC(mallinfo)().uordblks;
+  }
+#endif
+
+  if ( uordblks > 100000 )
+  { PL_set_prolog_flag("malloc", PL_ATOM, "ptmalloc");
+    is_ptmalloc = TRUE;
+  }
+
+  WEAK_IMPORT(malloc_trim); /* hope to have trim but it doesn't change the test */
 
   return is_ptmalloc;
 }
-#else
-#define is_ptmalloc FALSE
-#define initPTMalloc() FALSE
-#define fmalloc_trim(pad) (void)0
-#endif
 
 
 int
@@ -1908,10 +1891,10 @@ initMalloc(void)
 
 static
 PRED_IMPL("trim_heap", 0, trim_heap, 0)
-{ if ( is_tcmalloc && fMallocExtension_ReleaseFreeMemory )
-    fMallocExtension_ReleaseFreeMemory();
-  else if ( is_ptmalloc )
-    fmalloc_trim(0);
+{ if ( is_tcmalloc )
+    WEAK_TRY_CALL(MallocExtension_ReleaseFreeMemory);
+  else
+    WEAK_TRY_CALL(malloc_trim, 0);
 
   return TRUE;
 }
@@ -1931,23 +1914,18 @@ PRED_IMPL("thread_idle", 2, thread_idle, PL_FA_TRANSPARENT)
     return FALSE;
 
   if ( how == ATOM_short )
-  { trimStacks(TRUE PASS_LD);
-    if ( fMallocExtension_MarkThreadTemporarilyIdle &&
-	 fMallocExtension_MarkThreadBusy )
-      fMallocExtension_MarkThreadTemporarilyIdle();
+  { trimStacks(TRUE);
+    WEAK_TRY_CALL(MallocExtension_MarkThreadTemporarilyIdle);
   } else if ( how == ATOM_long )
   { LD->trim_stack_requested = TRUE;
     garbageCollect(GC_USER);
     LD->trim_stack_requested = FALSE;
-    if ( fMallocExtension_MarkThreadIdle  &&
-	 fMallocExtension_MarkThreadBusy )
-      fMallocExtension_MarkThreadIdle();
+    WEAK_TRY_CALL(MallocExtension_MarkThreadIdle);
   }
 
   rc = callProlog(NULL, A1, PL_Q_PASS_EXCEPTION, NULL);
 
-  if ( fMallocExtension_MarkThreadBusy )
-    fMallocExtension_MarkThreadBusy();
+  WEAK_TRY_CALL(MallocExtension_MarkThreadBusy);
 
   return rc;
 }
