@@ -110,7 +110,8 @@ struct PL_global_data
   } resources;
 
   struct
-  { sig_handler handlers[MAXSIGNAL];	/* How Prolog preceives signals */
+  { sig_handler handlers[NUM_SIGNALS];	/* How Prolog receives signals */
+#ifdef SIG_ALERT
     int		sig_alert;		/* our alert signal */
 #endif
   } signals;
@@ -451,13 +452,16 @@ struct PL_local_data
   struct PL_local_data *next_free;	/* see maybe_free_local_data() */
 
   struct
-  { int		pending[2];		/* PL_raise() pending signals */
+  { wsigmask_t	pending;		/* PL_raise() pending signals */
+#if STDC_CV_ALERT /* use C11 condition variable/mutex for thread signalling */
+    cnd_t	alert_cv;		/* notify when a signal is in pending */
+    mtx_t	alert_mtx;		/* lock on this while waiting for _cv */
+#endif
     int		current;		/* currently processing signal */
     int		is_sync;		/* current signal is synchronous */
 #ifndef __unix__
     int		forced;			/* Forced signal */
 #endif
-#ifdef HAVE_SIGNAL
 #ifdef O_C_STACK_GUARDED
     jmp_buf	context;		/* Guarded for C-stack overflows */
     int		sig_critical;		/* context is valid */
