@@ -131,9 +131,16 @@
  */
 #define COMMON(type) type
 
-#include "pl-builtin.h"
 #include "pl-macros.h"
 
+/* C11 gives us the _Static_assert operator, let's make it a little nicer.
+ * Accept either static_assert(cond, "message") or static_assertion(cond). */
+#define static_assert(condition, message) _Static_assert(condition, message)
+#define static_assertion(condition) _Static_assert(condition, "Assertion failed: ("#condition") [expansion: " A_STRINGIFY(condition) "]")
+
+#include "pl-builtin.h"
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 		      PROLOG SYSTEM OPTIONS
 
@@ -198,6 +205,11 @@ handy for it someone wants to add a data type to the system.
       Default for the `prefer_rationals` flag.
   O_RATIONAL_SYNTAX
       Default support for rational syntax (RAT_NATURAL or RAT_COMPAT)
+  O_SIGNALS
+      Include OS-level signal-handling code. This does not affect any
+      virtual Prolog "signals" defined above SIG_PROLOG_OFFSET, but
+      PL_thread_raise loses its "interrupt a blocking system call"
+      semantics when this is disabled.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define O_COMPILE_OR		1
@@ -350,9 +362,33 @@ typedef _sigset_t sigset_t;
 #include <stdarg.h>
 #include <limits.h>
 
-#ifdef HAVE_SIGNAL
+#ifdef HAVE_SIGNAL_H 
 #include <signal.h>
 #endif
+
+/* We have two important feature test macros relating to signals. The first is
+ * O_SIGNALS, which controls whether Prolog itself should perform OS-level
+ * signal handling, and is set by the USE_SIGNALS config directive.
+ * 
+ * The other macro is HAVE_OS_SIGNALS. This controls whether OS-level signal
+ * handling is available at all, e.g. from the on_signal/3 predicate.
+ */
+#ifndef HAVE_OS_SIGNALS
+# if defined(HAVE_SIGNAL) || defined(HAVE_SIGACTION)
+/* if we have at least one of these functions, assume we have access to
+ * OS-level signals, unless HAVE_OS_SIGNALS has been explicitly disabled. */
+#  define HAVE_OS_SIGNALS 1
+# endif
+#elif !HAVE_OS_SIGNALS
+/* If HAVE_OS_SIGNALS was explicitly defined to 0, turn that into an undef
+ * so that we can use either #if or #ifdef with it in libswipl code. */
+# undef HAVE_OS_SIGNALS
+#endif
+
+#if !HAVE_OS_SIGNALS
+/* If we can't access OS signals, don't include code that uses them. */
+# undef O_SIGNALS
+#endif
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
 #else
@@ -405,6 +441,17 @@ typedef _sigset_t sigset_t;
 #define MAXPATHLEN 1024
 #endif
 
+/* If we have a threads-supporting C11 environment, we can use C11 thread
+ * primitives (as a fallback, if we don't have a system-specific impl).
+ * We do not need this stuff on Windows as it is used for internal signal
+ * handling if no true signals are available.
+ */
+#if __STDC_VERSION__ >= 201112L && !defined(__STDC_NO_THREADS__) && !defined(__WINDOWS__)
+# include <threads.h>
+# undef thread_local /* we use this as an identifier, it's not in the C11 spec anyway */
+# define HAVE_STDC_THREADS 	1
+#endif
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 A common basis for C keywords.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -616,6 +663,7 @@ Magic for assertions.
 
 #define StackMagic(n)	((n) | 0x98765000)
 #define QID_MAGIC	StackMagic(1)	/* Query frame */
+#define QID_CMAGIC	StackMagic(2)	/* Closed Query frame */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 			  PROLOG DATA REPRESENTATION
@@ -1721,6 +1769,11 @@ typedef struct exception_frame		/* PL_throw exception environments */
 } exception_frame;
 
 
+typedef struct queryRef
+{ PL_engine_t	engine;			/* Engine for the query */
+  uintptr_t	offset;			/* queryFrane offset in local stack */
+} *QueryRef;
+
 #define QF_NODEBUG		0x0001	/* debug-able query */
 #define QF_DETERMINISTIC	0x0002	/* deterministic success */
 #define	QF_INTERACTIVE		0x0004	/* interactive goal (prolog()) */
@@ -1752,6 +1805,7 @@ struct queryFrame
   Choice	saved_bfr;		/* Saved choice-point */
   LocalFrame	saved_ltop;		/* Saved lTop */
   QueryFrame	parent;			/* Parent queryFrame */
+  QueryRef	qid;			/* Opaque query id */
   struct choice	choice;			/* First (dummy) choice-point */
   LocalFrame	saved_environment;	/* Parent local-frame */
 					/* Do not put anything between */
@@ -2054,8 +2108,7 @@ Temporary store/restore pointers to make them safe over GC/shift
 			   } while(0)
 
 
-#define QueryFromQid(qid)	((QueryFrame) valTermRef(qid))
-#define QidFromQuery(f)		(consTermRef(f))
+#define QidFromQuery(f)		((f)->qid)
 #define QID_EXPORT_WAM_TABLE	(qid_t)(-1)
 
 
@@ -2069,16 +2122,21 @@ still have signals that trigger Prolog   housekeeping  events. These are
 not bound to operating system signal handling though.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if HAVE_SIGNAL
-#define MAXSIGNAL		64	/* highest system signal number */
-#define SIG_PROLOG_OFFSET	32	/* Start of Prolog signals */
+#define SIG_PROLOG_OFFSET	32	/* Start of Prolog signals - specified in docs */
 
-#else /* HAVE_SIGNAL */
+#if HAVE_OS_SIGNALS
 
-#define MAXSIGNAL		32	/* highest system signal number */
-#define SIG_PROLOG_OFFSET	1	/* Start of Prolog signals */
+#define MINSIGNAL		1	/* number of first signal */
+#define MAXSIGNAL		64	/* highest supported signal number */
 
-#endif /* HAVE_SIGNAL */
+#else /* HAVE_OS_SIGNALS */
+
+#define MINSIGNAL		SIG_PROLOG_OFFSET /* No system signals, only Prolog sigs */
+#define MAXSIGNAL		(SIG_PROLOG_OFFSET+31)	/* we'd like this to fit in 32 bits */
+
+#endif /* HAVE_OS_SIGNALS */
+
+#define NUM_SIGNALS (MAXSIGNAL - MINSIGNAL + 1)
 
 #ifndef RETSIGTYPE
 #define RETSIGTYPE void
@@ -2092,17 +2150,83 @@ typedef struct
   int	      flags;			/* PLSIG_*, defined in pl-setup.c */
 } sig_handler, *SigHandler;
 
+/* Declare numbers for the virtual signals, in their own domain. For now, these
+ * just map to the numbers starting at SIG_PROLOG_OFFSET. Note, we could omit
+ * some of these definitions based on preprocessor statements, but there are
+ * few enough of them that it doesn't seem necessary.
+ */
+typedef enum virtual_signum
+{ VSIG_ATOM_GC,
+  VSIG_GC,
+  VSIG_THREAD_SIGNAL,
+  VSIG_CLAUSE_GC,
+  VSIG_PLABORT,
+  VSIG_TUNE_GC,
+  VSIG_MAX
+} virtual_signum;
 
+#define NUM_VSIGS 6 /* Preprocessor can see this constant */
+static_assertion(NUM_VSIGS == VSIG_MAX); /* Make sure it matches the enum */
+static_assertion(NUM_SIGNALS >= VSIG_MAX && NUM_SIGNALS < 128); /* Sanity check, 128 is arbitrary */
+static_assertion(SIG_PROLOG_OFFSET >= MINSIGNAL && SIG_PROLOG_OFFSET + NUM_VSIGS <= MAXSIGNAL);
+
+/* Define the macros that map vsignals into the signal domain. */
 #ifdef O_ATOMGC
-#define SIG_ATOM_GC	  (SIG_PROLOG_OFFSET+0)
+#define SIG_ATOM_GC	  (SIG_PROLOG_OFFSET+VSIG_ATOM_GC)
 #endif
-#define SIG_GC		  (SIG_PROLOG_OFFSET+1)
+#define SIG_GC		  (SIG_PROLOG_OFFSET+VSIG_GC)
 #ifdef O_PLMT
-#define SIG_THREAD_SIGNAL (SIG_PROLOG_OFFSET+2)
+#define SIG_THREAD_SIGNAL (SIG_PROLOG_OFFSET+VSIG_THREAD_SIGNAL)
 #endif
-#define SIG_CLAUSE_GC	  (SIG_PROLOG_OFFSET+3)
-#define SIG_PLABORT	  (SIG_PROLOG_OFFSET+4)
-#define SIG_TUNE_GC	  (SIG_PROLOG_OFFSET+5)
+#define SIG_CLAUSE_GC	  (SIG_PROLOG_OFFSET+VSIG_CLAUSE_GC)
+#define SIG_PLABORT	  (SIG_PROLOG_OFFSET+VSIG_PLABORT)
+#define SIG_TUNE_GC	  (SIG_PROLOG_OFFSET+VSIG_TUNE_GC)
+
+/* The "search for a free signal" functionality of PL_sigaction starts after
+ * the predefined VSIG numbers */
+#define SIG_USER_OFFSET	  (SIG_PROLOG_OFFSET+VSIG_MAX)
+
+/* Get a zero-based array index for this signal */
+#define SIGNAL_INDEX(sig)	((sig) - MINSIGNAL)
+/* Return the signal given an array index */
+#define SIGNAL_FROM_INDEX(idx)	((idx) + MINSIGNAL)
+/* Is this a valid signal number? */
+#define IS_VALID_SIGNAL(sig)	((sig) >= MINSIGNAL && (sig) <= MAXSIGNAL)
+/* Is this a virtual signal? */
+#define IS_VSIG(sig)		((sig) >= SIG_PROLOG_OFFSET)
+
+#if O_DEBUG
+#undef SIGNAL_INDEX
+static inline int
+SIGNAL_INDEX(int sig)
+{ assert(IS_VALID_SIGNAL(sig));
+  return sig - MINSIGNAL;
+}
+#endif
+
+/* We want fast types for signal bitmasks; on a 64-bit arch this is probably the same as uint64_t */
+typedef uint_fast32_t		sigmask_t;
+
+/* How many bits can fit in a single sigmask_t? */
+#define SIGMASK_WIDTH		(sizeof(sigmask_t) * 8)
+/* How many sigmask_t's does it take to store all supported signals? */
+#define SIGMASK_WORDS		((NUM_SIGNALS + SIGMASK_WIDTH - 1) / SIGMASK_WIDTH)
+/* Which sigmask word is this signal in? */
+#define SIGMASK_WORD(sig)	(SIGNAL_INDEX(sig) / SIGMASK_WIDTH)
+/* Which bit of the given word? */
+#define SIGMASK_BIT(sig)	(SIGNAL_INDEX(sig) % SIGMASK_WIDTH)
+/* Single-bit mask corresponding to SIGMASK_BIT */
+#define SIGMASK_MASK(sig)	(((sigmask_t)1) << SIGMASK_BIT(sig))
+
+/* Obtain signal given sigmask word and bit */
+#define SIGNAL_FROM_SIGMASK(word, bit) SIGNAL_FROM_INDEX((word) * SIGMASK_WIDTH + (bit))
+
+typedef sigmask_t		wsigmask_t[SIGMASK_WORDS];
+
+/* Helper macros for interacting with sigmask_t arrays */
+#define WSIGMASK_ISSET(wm, sig)	(((wm)[SIGMASK_WORD(sig)] & SIGMASK_MASK(sig)) != 0)
+#define WSIGMASK_SET(wm, sig)	ATOMIC_OR(&(wm)[SIGMASK_WORD(sig)], SIGMASK_MASK(sig))
+#define WSIGMASK_CLEAR(wm, sig)	ATOMIC_AND(&(wm)[SIGMASK_WORD(sig)], ~SIGMASK_MASK(sig))
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Guard against C-stack overflows. This is   done  for POSIX systems using
@@ -2114,7 +2238,7 @@ simpler, but a lot slower as it  implies   a  system  call. We assume no
 other signals are involved and unblock SIGSEGV by hand.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if defined(HAVE_SIGALTSTACK) && !defined(__SANITIZE_ADDRESS__)
+#if O_SIGNALS && defined(HAVE_SIGALTSTACK) && !defined(__SANITIZE_ADDRESS__)
 #define O_C_STACK_GUARDED 1
 #define C_STACK_OVERFLOW_GUARDED(rc, code, cleanup) \
 	do						\
@@ -2662,6 +2786,37 @@ with any luck LTO could address that.
 # define ALLOC_INLINE
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Helpers for weak function references, where we want our behavior to change
+based on if we've been linked against a specific library or not. To use a
+weak function, you should (e.g. for weak-importing malloc):
+
+1. WEAK_DECLARE(void *, malloc, (size_t size)); at .c file scope
+2. WEAK_IMPORT(malloc) in startup code (returns success)
+3. if (WEAK_FUNC(malloc)) to check if the import succeeded
+4a. WEAK_FUNC(malloc)(sizeof(x)) to call the function, if import worked, or
+4b. WEAK_TRY_CALL(malloc, sizeof(x)) to combine check and call.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static inline int __ptr_to_bool(const void *ptr) { return ptr != NULL; }
+#if HAVE_WEAK_ATTRIBUTE
+# define WEAK_DECLARE(RType, Name, Params) \
+	extern RType __attribute__((weak)) Name Params
+# define WEAK_IMPORT(Name) __ptr_to_bool(&Name)
+# define WEAK_FUNC(Name) (&Name)
+#elif defined(HAVE_DLOPEN) || defined(HAVE_SHL_LOAD) || defined(EMULATE_DLOPEN)
+# define WEAK_DECLARE(RType, Name, Params) \
+	static RType (*wf##Name) Params = NULL
+# define WEAK_IMPORT(Name) __ptr_to_bool((wf##Name = PL_dlsym(NULL, #Name)))
+# define WEAK_FUNC(Name) (wf##Name)
+#else
+# define WEAK_DECLARE(RType, Name, Params) \
+	extern RType Name Params
+# define WEAK_IMPORT(Name) 0
+# define WEAK_FUNC(Name) 0
+#endif
+#define WEAK_TRY_CALL(Name, ...) (WEAK_FUNC(Name) != NULL ? WEAK_FUNC(Name)(__VA_ARGS__) : 0)
+
 #include "pl-util.h"			/* (Debug) utilities */
 #include "pl-alloc.h"			/* Allocation primitives */
 #include "pl-init.h"			/* Declarations needed by pl-init.c */
