@@ -65,6 +65,7 @@ option  parsing,  initialisation  and  handling  of errors and warnings.
 #include "os/pl-prologflag.h"
 #include "os/pl-ctype.h"
 #include "os/pl-utf8.h"
+#include "os/pl-cstack.h"
 #include <errno.h>
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
@@ -490,7 +491,11 @@ initPaths(int argc, const char **argv)
 #endif
   }
 
-  systemDefaults.startup = store_string(PrologPath(DEFSTARTUP, plp, sizeof(plp)));
+#ifdef DEFSTARTUP
+  systemDefaults.startup = store_string(DEFSTARTUP);
+#else
+  systemDefaults.startup = NULL;
+#endif
 
 #ifdef O_XOS
   if ( systemDefaults.home )
@@ -648,7 +653,7 @@ initDefaultOptions(void)
   GD->options.scriptFiles      = NULL;
   GD->options.saveclass	       = store_string("none");
   GD->options.on_error	       = store_string("print");
-  GD->options.on_warning       = store_string("warning");
+  GD->options.on_warning       = store_string("print");
 
   if ( systemDefaults.goal )
     opt_append(&GD->options.goals, systemDefaults.goal);
@@ -777,7 +782,7 @@ parseCommandLineOptions(int argc0, char **argv0, char **argvleft, int compile)
 #endif
       } else if ( (optval=is_longopt(s, "traditional")) )
       { setTraditional();
-#ifdef HAVE_SIGNAL
+#if O_SIGNALS && defined(SIG_ALERT)
       } else if ( (optval=is_longopt(s, "sigalert")) )
       { char *e;
 	long sig = strtol(optval, &e, 10);
@@ -1069,7 +1074,7 @@ PL_initialise(int argc, char **argv)
   initPaths(argc, (const char**)argv);	/* fetch some useful paths */
 
   { GET_LD
-#ifdef HAVE_SIGNAL
+#ifdef O_SIGNALS
     setPrologFlagMask(PLFLAG_SIGNALS);	/* default: handle signals */
 #endif
 
@@ -1675,14 +1680,20 @@ vsysError(const char *fm, va_list args)
     unblockSignals(&LD->gc.saved_sigmask);
   }
 
+  Sfprintf(Serror, "\n");
+  save_backtrace("SYSERROR");
+  print_backtrace_named("SYSERROR");
+
 #if defined(O_DEBUGGER)
   Sfprintf(Serror, "\n\nPROLOG STACK:\n");
   PL_backtrace(10, 0);
   Sfprintf(Serror, "]\n");
 #endif /*O_DEBUGGER*/
 
-  if ( GD->bootsession )
-    PL_abort_process();
+  if ( !(true(Sinput, SIO_ISATTY) &&
+	 true(Serror, SIO_ISATTY)) ||
+       GD->bootsession )
+    PL_abort_process();			/* non-interactive or booting */
 
 action:
 #ifdef HAVE_GETPID
