@@ -82,9 +82,16 @@ access.   Finally  it holds the code to handle signals transparently for
 foreign language code or packages with which Prolog was linked together.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#if USE_LD_MACROS
+#define allocStacks(_)		LDFUNC(allocStacks, _)
+#define initSignals(_)		LDFUNC(initSignals, _)
+#endif
+
+#define LDFUNC_DECLARATIONS
 static int allocStacks(void);
 static void initSignals(void);
 static void gcPolicy(Stack s, int policy);
+#undef LDFUNC_DECLARATIONS
 
 int
 setupProlog(void)
@@ -97,8 +104,8 @@ setupProlog(void)
 
   LD->critical = 0;
   LD->magic = LD_MAGIC;
-  LD->signal.pending[0] = 0;
-  LD->signal.pending[1] = 0;
+  for (int i = 0; i < SIGMASK_WORDS; i++)
+    LD->signal.pending[i] = 0;
   LD->statistics.start_time = WallTime();
 
   startCritical;
@@ -191,6 +198,11 @@ initPrologLocalData(DECL_LD)
   LD->transaction.gen_base = GEN_INFINITE;
 #endif
 
+#if STDC_CV_ALERT
+  cnd_init(&LD->signal.alert_cv);
+  mtx_init(&LD->signal.alert_mtx, mtx_plain);
+#endif
+
   updateAlerted(LD);
 }
 
@@ -222,7 +234,11 @@ they  define  signal handlers to be int functions.  This should be fixed
 some day.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define PLSIG_USERFLAGS 0x0000ffff	/* range of API-visible flags */
+#define PLSIG_STATEFLAGS 0xffff0000	/* range of internal flags */
+
 #define PLSIG_PREPARED 0x00010000	/* signal is prepared */
+#define PLSIG_IGNORED  0x00020000	/* signal is ignored */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Define the signals and  their  properties.   This  could  be  nicer, but
@@ -236,7 +252,7 @@ static struct signame
   int	      flags;
 } signames[] =
 {
-#ifdef HAVE_SIGNAL
+#ifdef HAVE_OS_SIGNALS
 #ifdef SIGHUP
   { SIGHUP,	"hup",    0},
 #endif
@@ -252,7 +268,7 @@ static struct signame
 #endif
   { SIGSEGV,	"segv",   0},
 #ifdef SIGPIPE
-  { SIGPIPE,	"pipe",   0},
+  { SIGPIPE,	"pipe",   PLSIG_IGNORE},
 #endif
 #ifdef SIGALRM
   { SIGALRM,	"alrm",   PLSIG_THROW},
@@ -315,7 +331,7 @@ static struct signame
 #ifdef SIGPWR
   { SIGPWR,	"pwr",    0},
 #endif
-#endif /*HAVE_SIGNAL*/
+#endif /*HAVE_OS_SIGNALS*/
 
 /* The signals below here are recorded as Prolog interrupts, but
    not supported by OS signals.  They start at offset 32.
@@ -451,7 +467,7 @@ is_fatal_signal(int sig)
 void
 dispatch_signal(int sig, int sync)
 { GET_LD
-  SigHandler sh = &GD->signals.handlers[sig-1];
+  SigHandler sh = &GD->signals.handlers[SIGNAL_INDEX(sig)];
   fid_t fid;
   term_t lTopSave;
   int saved_current_signal;
@@ -494,14 +510,16 @@ dispatch_signal(int sig, int sync)
   if ( is_fatal_signal(sig) && sig == LD->signal.current )
     sysError("Recursively received fatal signal %d", sig);
 
-  if ( gc_status.active && sig < SIG_PROLOG_OFFSET )
+  if ( gc_status.active && !IS_VSIG(sig) )
   { fatalError("Received signal %d (%s) while in %ld-th garbage collection",
 	       sig, signal_name(sig), LD->gc.stats.totals.collections);
   }
 
-  if ( (LD->critical || (true(sh, PLSIG_SYNC) && !sync)) &&
-       !is_fatal_signal(sig) &&
-       sig != SIGINT )
+  if ( (LD->critical || (true(sh, PLSIG_SYNC) && !sync))
+#if O_SIGNALS
+       && sig != SIGINT
+#endif
+       && !is_fatal_signal(sig)	)
   { PL_raise(sig);			/* wait for better times! */
     return;
   }
@@ -565,8 +583,8 @@ dispatch_signal(int sig, int sync)
 #endif
 
     DEBUG(MSG_SIGNAL,
-	  Sdprintf("Handler %p finished (pending=0x%x,0x%x)\n",
-		   sh->handler, LD->signal.pending[0], LD->signal.pending[1]));
+	  Sdprintf("Handler %p finished (pending=0x%"PRIxFAST32",0x%"PRIxFAST32")\n",
+		   sh->handler, LD->signal.pending[0], SIGMASK_WORDS > 1 ? LD->signal.pending[1] : 0));
 
     if ( !ex_pending && exception_term && !sync )	/* handler: PL_raise_exception() */
       fatalError("Async exception handler for signal %s (%d) raised "
@@ -641,13 +659,21 @@ set_sighandler(int sig, handler_t func)
 }
 
 static SigHandler
-prepareSignal(int sig)
-{ SigHandler sh = &GD->signals.handlers[sig-1];
+prepareSignal(int sig, int plsig_flags)
+{ SigHandler sh = &GD->signals.handlers[SIGNAL_INDEX(sig)];
+  int current_state = sh->flags & PLSIG_STATEFLAGS;
+  int desired_state = (plsig_flags & PLSIG_IGNORE) ? PLSIG_IGNORED : PLSIG_PREPARED;
 
-  if ( false(sh, PLSIG_PREPARED) )
-  { set(sh, PLSIG_PREPARED);
-    if ( sig < SIG_PROLOG_OFFSET )
-      sh->saved_handler = set_sighandler(sig, pl_signal_handler);
+  plsig_flags &= ~(PLSIG_STATEFLAGS | PLSIG_IGNORE);
+
+  if ( current_state != desired_state )
+  { clearFlags(sh);
+    set(sh, desired_state | plsig_flags);
+    if ( !IS_VSIG(sig) )
+    { handler_t old_handler = set_sighandler(sig, desired_state == PLSIG_IGNORED ? SIG_IGN : pl_signal_handler);
+      if ( current_state == 0 )
+        sh->saved_handler = old_handler;
+    }
   }
 
   return sh;
@@ -656,10 +682,10 @@ prepareSignal(int sig)
 
 static void
 unprepareSignal(int sig)
-{ SigHandler sh = &GD->signals.handlers[sig-1];
+{ SigHandler sh = &GD->signals.handlers[SIGNAL_INDEX(sig)];
 
-  if ( true(sh, PLSIG_PREPARED) )
-  { if ( sig < SIG_PROLOG_OFFSET )
+  if ( true(sh, PLSIG_STATEFLAGS) )
+  { if ( !IS_VSIG(sig) )
       set_sighandler(sig, sh->saved_handler);
     sh->flags         = 0;
     sh->handler       = NULL;
@@ -669,6 +695,8 @@ unprepareSignal(int sig)
 }
 
 
+#if O_SIGNALS
+
 #ifdef SIGHUP
 static void
 hupHandler(int sig)
@@ -679,7 +707,6 @@ hupHandler(int sig)
 #endif
 
 
-#ifdef HAVE_SIGNAL
 /* terminate_handler() is called on termination signals like SIGTERM.
    It runs hooks registered using PL_exit_hook() and then kills itself.
    The hooks are called with the exit status `3`.
@@ -732,7 +759,7 @@ initTerminationSignals(void)
   terminate_on_signal(SIGQUIT);
 #endif
 }
-#endif /*HAVE_SIGNAL*/
+#endif /*O_SIGNALS*/
 
 #ifdef O_C_STACK_GUARDED
 static void
@@ -836,7 +863,7 @@ with EINTR and thus make them interruptable for thread-signals.
 #ifdef SIG_ALERT
 static void
 alert_handler(int sig)
-{ SigHandler sh = &GD->signals.handlers[sig-1];
+{ SigHandler sh = &GD->signals.handlers[SIGNAL_INDEX(sig)];
 
   if ( sh->saved_handler &&
        sh->saved_handler != SIG_IGN &&
@@ -847,19 +874,16 @@ alert_handler(int sig)
 
 
 static void
-initSignals(void)
-{ GET_LD
-
+initSignals(DECL_LD)
+{
+#if O_SIGNALS
   /* This is general signal handling that is not strictly needed */
   if ( truePrologFlag(PLFLAG_SIGNALS) )
   { struct signame *sn = signames;
-#ifdef HAVE_SIGNAL
-#ifdef SIGPIPE
-    set_sighandler(SIGPIPE, SIG_IGN);
-#endif
+#ifdef HAVE_OS_SIGNALS
     initTerminationSignals();
     initGuardCStack();
-#endif /*HAVE_SIGNAL*/
+#endif /*HAVE_OS_SIGNALS*/
     initBackTrace();
     for( ; sn->name; sn++)
     {
@@ -869,9 +893,7 @@ initSignals(void)
 	sn->flags = 0;
 #endif
       if ( sn->flags )
-      { SigHandler sh = prepareSignal(sn->sig);
-	sh->flags |= sn->flags;
-      }
+        prepareSignal(sn->sig, sn->flags);
     }
 
 #ifdef SIGHUP
@@ -886,6 +908,7 @@ initSignals(void)
   if ( GD->signals.sig_alert )
     PL_signal(GD->signals.sig_alert|PL_SIGNOFRAME, alert_handler);
 #endif
+#endif /*O_SIGNALS*/
 
   /* these signals are not related to Unix signals and can thus */
   /* be enabled always */
@@ -917,8 +940,8 @@ resetSignals(void)
 { GET_LD
 
   LD->signal.current = 0;
-  LD->signal.pending[0] = 0;
-  LD->signal.pending[1] = 0;
+  for (int i = 0; i < SIGMASK_WORDS; i++)
+    LD->signal.pending[i] = 0;
 }
 
 #if defined(O_PLMT) && defined(HAVE_PTHREAD_SIGMASK)
@@ -929,7 +952,7 @@ resetSignals(void)
 #define sigprocmask(how, new, old) pthread_sigmask(how, new, old)
 #endif
 
-#ifdef HAVE_SIGPROCMASK
+#if O_SIGNALS && defined(HAVE_SIGPROCMASK)
 
 void
 allSignalMask(sigset_t *set)
@@ -1028,7 +1051,7 @@ blockSignal(int sig)
   DEBUG(1, Sdprintf("signal %d\n", sig));
 }
 
-#else /*HAVE_SIGPROCMASK*/
+#else /*O_SIGNALS && defined(HAVE_SIGPROCMASK)*/
 
 void blockSignals(sigset_t *old) {}
 void unblockSignals(sigset_t *old) {}
@@ -1048,14 +1071,14 @@ int
 PL_sigaction(int sig, pl_sigaction_t *act, pl_sigaction_t *old)
 { SigHandler sh = NULL;
 
-  if ( sig < 0 || sig > MAXSIGNAL )
+  if ( sig && !IS_VALID_SIGNAL(sig) )
   { errno = EINVAL;
     return -1;
   }
 
   if ( sig == 0 )
-  { for(sig=SIG_PROLOG_OFFSET; sig<MAXSIGNAL; sig++)
-    { sh = &GD->signals.handlers[sig-1];
+  { for(sig=SIG_USER_OFFSET; sig<=MAXSIGNAL; sig++)
+    { sh = &GD->signals.handlers[SIGNAL_INDEX(sig)];
       if ( sh->flags == 0 )
 	break;
     }
@@ -1064,7 +1087,7 @@ PL_sigaction(int sig, pl_sigaction_t *act, pl_sigaction_t *old)
       return -2;
     }
   } else
-  { sh = &GD->signals.handlers[sig-1];
+  { sh = &GD->signals.handlers[SIGNAL_INDEX(sig)];
   }
 
   if ( old )
@@ -1092,9 +1115,8 @@ PL_sigaction(int sig, pl_sigaction_t *act, pl_sigaction_t *old)
     if ( active )
     { sh->handler   = act->sa_cfunction;
       sh->predicate = act->sa_predicate;
-      sh->flags     = (sh->flags&~0xffff)|act->sa_flags;
-      if ( false(sh, PLSIG_PREPARED) )
-	prepareSignal(sig);
+      sh->flags     = (sh->flags&~PLSIG_USERFLAGS)|act->sa_flags;
+      prepareSignal(sig, act->sa_flags);
     } else
     { unprepareSignal(sig);
       sh->handler   = NULL;
@@ -1121,7 +1143,7 @@ PL_signal(int sigandflags, handler_t func)
   if ( (sigandflags&PL_SIGNOFRAME) )
     act.sa_flags |= PLSIG_NOFRAME;
 
-  if ( PL_sigaction((sigandflags & 0xffff), &act, &old) >= 0 )
+  if ( PL_sigaction((sigandflags & PLSIG_USERFLAGS), &act, &old) >= 0 )
   { if ( (old.sa_flags&PLSIG_PREPARED) && old.sa_cfunction )
       return old.sa_cfunction;
 
@@ -1149,10 +1171,8 @@ PL_handle_signals(void)
 #define handleSigInt(_) LDFUNC(handleSigInt, _)
 static int
 handleSigInt(DECL_LD)
-{ int intmask = 1<<(SIGINT-1);
-
-  if ( LD->signal.forced == SIGINT && LD->signal.pending[0] & intmask )
-  { ATOMIC_AND(&LD->signal.pending[0], ~intmask);
+{ if ( LD->signal.forced == SIGINT && WSIGMASK_ISSET(LD->signal.pending, SIGINT) )
+  { WSIGMASK_CLEAR(LD->signal.pending, SIGINT);
 
     LD->signal.forced = 0;
     dispatch_signal(SIGINT, TRUE);
@@ -1186,10 +1206,10 @@ handleSignals(DECL_LD)
   if ( LD->critical )
     return 0;
 
-  for(i=0; i<2; i++)
+  for(i=0; i<SIGMASK_WORDS; i++)
   { while( LD->signal.pending[i] )
-    { int sig = 1+32*i;
-      unsigned mask = 1;
+    { int sig = MINSIGNAL+SIGMASK_WIDTH*i;
+      sigmask_t mask = 1;
 
       for( ; mask ; mask <<= 1, sig++ )
       { if ( LD->signal.pending[i] & mask )
@@ -1260,11 +1280,13 @@ f_endCritical(DECL_LD)
 }
 
 
-#ifdef HAVE_SIGNAL
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 on_signal(?SigNum, ?SigName, :OldHandler, :NewHandler)
 
 Assign NewHandler to be called if signal arrives.
+
+We always support this even when compiled without OS-level signal support,
+because of internal virtual signal handling.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
@@ -1314,7 +1336,7 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
        !get_meta_arg(new, mnew, new) )
     return FALSE;
 
-  if ( PL_get_integer(sig, &sign) && sign >= 1 && sign <= MAXSIGNAL )
+  if ( PL_get_integer(sig, &sign) && IS_VALID_SIGNAL(sign) )
   { TRY(PL_unify_atom_chars(name, signal_name(sign)));
   } else if ( PL_get_atom_chars(name, &sn) )
   { if ( (sign = signal_index(sn)) != -1 )
@@ -1324,10 +1346,12 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
   } else
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_signal, sig);
 
-  sh = &GD->signals.handlers[sign-1];
+  sh = &GD->signals.handlers[SIGNAL_INDEX(sign)];
 
-  if ( false(sh, PLSIG_PREPARED) )		/* not handled */
+  if ( false(sh, PLSIG_STATEFLAGS) )		/* not handled */
   { TRY(PL_unify_atom(old, ATOM_default));
+  } else if ( true(sh, PLSIG_IGNORED) )		/* signal ignored */
+  { TRY(PL_unify_atom(old, ATOM_ignore));
   } else if ( true(sh, PLSIG_THROW) )		/* throw exception */
   { TRY(PL_unify_atom(old, ATOM_throw));
   } else if ( sh->predicate )			/* call predicate */
@@ -1359,15 +1383,15 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
   if ( PL_get_atom(new, &a) )
   { if ( a == ATOM_default )
     { unprepareSignal(sign);
+    } else if ( a == ATOM_ignore )
+    { prepareSignal(sign, PLSIG_IGNORE);	/* request to ignore this signal */
     } else if ( a == ATOM_throw )
-    { sh = prepareSignal(sign);
-      set(sh, PLSIG_THROW|PLSIG_SYNC);
+    { sh = prepareSignal(sign, PLSIG_THROW|PLSIG_SYNC);
       sh->handler   = NULL;
       sh->predicate = NULL;
     } else if ( a == ATOM_debug )
-    { sh = prepareSignal(sign);
+    { sh = prepareSignal(sign, 0);
 
-      clear(sh, PLSIG_THROW|PLSIG_SYNC);
       sh->handler = (handler_t)PL_interrupt;
       sh->predicate = NULL;
 
@@ -1379,9 +1403,7 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
 	return FALSE;
       pred = lookupProcedure(PL_new_functor(a, 1), m);
 
-      sh = prepareSignal(sign);
-      clear(sh, PLSIG_THROW);
-      set(sh, PLSIG_SYNC);
+      sh = prepareSignal(sign, PLSIG_SYNC);
       sh->handler = NULL;
       sh->predicate = pred;
     }
@@ -1392,8 +1414,7 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
     _PL_get_arg(1, new, a);
 
     if ( PL_get_pointer(a, &f) )
-    { sh = prepareSignal(sign);
-      clear(sh, PLSIG_THROW|PLSIG_SYNC);
+    { sh = prepareSignal(sign, 0);
       sh->handler = (handler_t)f;
       sh->predicate = NULL;
 
@@ -1407,8 +1428,6 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
   succeed;
 }
 
-#endif /*HAVE_SIGNAL*/
-
 
 		 /*******************************
 		 *	       STACKS		*
@@ -1536,9 +1555,8 @@ init_stack(Stack s, char *name, size_t size, size_t spare, int gc)
 
 
 static int
-allocStacks(void)
-{ GET_LD
-  size_t minglobal = 8*SIZEOF_VOIDP K;
+allocStacks(DECL_LD)
+{ size_t minglobal = 8*SIZEOF_VOIDP K;
   size_t minlocal  = 4*SIZEOF_VOIDP K;
   size_t mintrail  = 4*SIZEOF_VOIDP K;
   size_t minarg    = 1*SIZEOF_VOIDP K;
@@ -1746,6 +1764,28 @@ freePrologLocalData(PL_local_data_t *ld)
 #endif
 }
 
+/* The following definitions aren't necessary for compiling, and in fact
+ * you could comment this whole section out without breaking the code.
+ * However, they don't take up much space in the binary and they assist
+ * in C-level debugging, so I'm leaving them in regardless of O_DEBUG. */
+
+const intptr_t __PL_ld = -1;
+const intptr_t PL__ctx = -1;
+
+inline PL_local_data_t*
+(__FIND_LD)(PL_local_data_t *pl_ld, control_t pl_ctx, PL_local_data_t *fallback)
+{ if ((intptr_t)pl_ld != -1)
+  { return pl_ld; }
+  if ((intptr_t)pl_ctx != -1)
+  { return pl_ctx->engine; }
+  return fallback;
+}
+
+#ifndef no_local_ld
+PL_local_data_t*
+no_local_ld(void)
+{ return NULL; }
+#endif
 
 
 		 /*******************************
@@ -1866,10 +1906,8 @@ PRED_IMPL("$set_prolog_stack", 4, set_prolog_stack, 0)
 BeginPredDefs(setup)
   PRED_DEF("$set_prolog_stack",	  4, set_prolog_stack,	  0)
   PRED_DEF("trim_stacks",	  0, trim_stacks,	  0)
-#ifdef HAVE_SIGNAL
   PRED_DEF("$on_signal",	  4, on_signal,		  0)
 #ifdef SIG_ALERT
   PRED_DEF("prolog_alert_signal", 2, prolog_alert_signal, 0)
 #endif
-#endif
 EndPredDefs
