@@ -764,13 +764,13 @@ traceAction(char *cmd, int port, LocalFrame frame, Choice bfr,
 		{ FeedBack("No show context\n");
 		}
 		return ACTION_AGAIN;
-    case 'm':	FeedBack("Exception details");
+    case 'm':	FeedBack("Exception details\n");
 	        if ( port & EXCEPTION_PORT )
 		{ exceptionDetails();
 		} else
 		   Warn("No exception\n");
 		return ACTION_AGAIN;
-    case 'L':	FeedBack("Listing");
+    case 'L':	FeedBack("Listing\n");
 		listGoal(frame);
 		return ACTION_AGAIN;
     case 'S':	FeedBack("Save goal");
@@ -832,7 +832,8 @@ helpTrace(void)
     { "s",		   "skip over" },
     { "[level] S",	   "save goal [at level]" },
     { "u",		   "up (complete goal)" },
-    { "s",		   "(quoted) write goals" },
+    { "p",		   "print goals" },
+    { "w",		   "(quoted) write goals" },
     { "m",		   "exception details" },
     { "C",		   "toggle show context" },
   #if O_DEBUG
@@ -1105,7 +1106,7 @@ exceptionDetails()
   { int rc;
 
     Sflush(Suser_output);		/* make sure to stay in sync */
-    Sfprintf(Sdout, "\n\tException term: ");
+    Sfprintf(Sdout, "\tException term: ");
     rc = PL_write_term(Sdout, except, 1200, PL_WRT_QUOTED);
     Sfprintf(Sdout, "\n\t       Message: %s\n", messageToString(except));
 
@@ -1893,10 +1894,8 @@ PL_interrupt(int sig)
 
 
 void
-initTracer(void)
-{ GET_LD
-
-  debugstatus.visible      =
+initTracer(DECL_LD)
+{ debugstatus.visible      =
   debugstatus.leashing     = CALL_PORT|FAIL_PORT|REDO_PORT|EXIT_PORT|
 			     EXCEPTION_PORT;
   debugstatus.showContext  = FALSE;
@@ -1907,10 +1906,9 @@ initTracer(void)
 }
 
 int
-enable_debug_on_interrupt(int enable)
-{ GET_LD
-
-#ifdef SIGINT
+enable_debug_on_interrupt(DECL_LD int enable)
+{
+#if O_SIGNALS && defined(SIGINT)
   if ( enable )
   { if ( truePrologFlag(PLFLAG_SIGNALS) )
     { PL_signal(SIGINT, PL_interrupt);
@@ -2549,6 +2547,29 @@ in_clause_jump(Choice ch)
 }
 
 
+static atom_t
+choice_type_atom(Choice ch)
+{ static const atom_t types[] =
+  { ATOM_jump,
+    ATOM_clause,
+    ATOM_top,
+    ATOM_catch,
+    ATOM_debug
+  };
+
+  if ( ch->type == CHP_JUMP &&
+       in_clause_jump(ch) == (size_t)-1 )
+  { if ( ch->value.pc == SUPERVISOR(next_clause) )
+      return ATOM_clause;
+    if ( decode(ch->value.pc[0]) == I_FREDO )
+      return ATOM_foreign;
+    assert(0);
+    return FALSE;
+  } else
+    return types[ch->type];
+}
+
+
 static
 PRED_IMPL("prolog_choice_attribute", 3, prolog_choice_attribute, 0)
 { PRED_LD
@@ -2570,30 +2591,20 @@ PRED_IMPL("prolog_choice_attribute", 3, prolog_choice_attribute, 0)
   } else if ( key == ATOM_frame )
   { return PL_unify_frame(A3, ch->frame);
   } else if ( key == ATOM_type )
-  { static const atom_t types[] =
-    { ATOM_jump,
-      ATOM_clause,
-      ATOM_top,
-      ATOM_catch,
-      ATOM_debug
-    };
-
-    if ( ch->type == CHP_JUMP &&
-	 in_clause_jump(ch) == (size_t)-1 )
-    { if ( ch->value.pc == SUPERVISOR(next_clause) )
-	return PL_unify_atom(A3, ATOM_clause);
-      if ( decode(ch->value.pc[0]) == I_FREDO )
-	return PL_unify_atom(A3, ATOM_foreign);
-      assert(0);
-      return FALSE;
-    } else
-      return PL_unify_atom(A3, types[ch->type]);
+  { return PL_unify_atom(A3, choice_type_atom(ch));
   } else if ( key == ATOM_pc )
   { size_t offset = in_clause_jump(ch);
 
     if ( offset != (size_t)-1 )
       return PL_unify_int64(A3, offset);
     return FALSE;
+  } else if ( key == ATOM_clause )
+  { if ( ch->type == CHP_CLAUSE )
+    { return PL_unify_clref(A3, ch->value.clause.cref->value.clause);
+    } else if ( choice_type_atom(ch) == ATOM_clause )
+    { Sdprintf("S_ALLCLAUSES not yet supported\n");
+    }
+    return FALSE;
   } else
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_key, A2);
 
